if __name__ == '__main__':
    print('Tony has been updated and this way of running the program is deprecated. Please refer to the README for instructions: https://github.com/Pasu4/neuro-api-tony?tab=readme-ov-file#installation')
    input('\nPress Enter to exit...')


import random

import jsonschema._utils
import jsonschema.benchmarks
import jsonschema.exceptions
import jsonschema.tests
import wx
from jsf import JSF

from .api import *
from .model import NeuroAction, TonyModel
from .view import TonyView


def action_id_generator():
    '''Generate a unique ID for an action.'''

    i = 0
    while True:
        yield f'action_{i}'
        i += 1

class TonyController:

    def __init__(self, app: wx.App, log_level: str):
        self.app = app
        self.model = TonyModel()
        self.view = TonyView(app, self.model, log_level)
        self.api = NeuroAPI()

        self.active_actions_force: ActionsForceCommand | None = None

        self.id_generator = action_id_generator()

        self.inject()

    def run(self, address: str, port: int):
        # Schedule the API start to run after the main loop starts
        wx.CallAfter(self.api.start, address, port)

        self.view.show()
        self.app.MainLoop()

    def inject(self):
        '''Inject methods into the view and API.'''

        self.api.on_startup = self.on_startup
        self.api.on_context = self.on_context
        self.api.on_actions_register = self.on_actions_register
        self.api.on_actions_unregister = self.on_actions_unregister
        self.api.on_actions_force = self.on_actions_force
        self.api.on_action_result = self.on_action_result
        self.api.on_shutdown_ready = self.on_shutdown_ready
        self.api.on_unknown_command = self.on_unknown_command
        self.api.log_system = self.view.log_system
        self.api.log_debug = self.view.log_debug
        self.api.log_info = self.view.log_info
        self.api.log_warning = self.view.log_warning
        self.api.log_error = self.view.log_error
        self.api.log_critical = self.view.log_critical
        self.api.log_raw = self.view.log_raw
        self.api.get_delay = lambda: float(self.view.controls.latency / 1000)

        self.view.on_execute = self.on_view_execute
        self.view.on_delete_action = self.on_view_delete_action
        self.view.on_unlock = self.on_view_unlock
        self.view.on_clear_logs = self.on_view_clear_logs
        self.view.on_send_actions_reregister_all = self.on_view_send_actions_reregister_all
        self.view.on_send_shutdown_graceful = self.on_view_send_shutdown_graceful
        self.view.on_send_shutdown_graceful_cancel = self.on_view_send_shutdown_graceful_cancel
        self.view.on_send_shutdown_immediate = self.on_view_send_shutdown_immediate

    def on_any_command(self, cmd: Any):
        '''Callback for any command received from the API.'''

    def on_startup(self, cmd: StartupCommand):
        '''Handle the startup command.'''

        self.model.clear_actions()
        self.view.clear_actions()

    def on_context(self, cmd: ContextCommand):
        '''Handle the context command.'''

        self.view.log_context(cmd.message, silent=cmd.silent)

    def on_actions_register(self, cmd: ActionsRegisterCommand):
        '''Handle the actions/register command.'''

        for action in cmd.actions:

            # Check if an action with the same name already exists
            if self.model.has_action(action.name):
                self.view.log_warning(f'Action "{action.name}" already exists. Ignoring.')
                continue

            self.model.add_action(action)
            wx.CallAfter(self.view.add_action, action)
            self.view.log_system(f'Action registered: {action.name}')
            self.view.log_description(f'{action.name}: {action.description}')

    def on_actions_unregister(self, cmd: ActionsUnregisterCommand):
        '''Handle the actions/unregister command.'''

        for name in cmd.action_names:
            if not self.model.has_action(name):
                self.view.log_info(f'Action "{name}" does not exist.')

            self.model.remove_action_by_name(name)
            self.view.remove_action_by_name(name)
            self.view.log_system(f'Action unregistered: {name}')

    def on_actions_force(self, cmd: ActionsForceCommand):
        '''Handle the actions/force command.'''

        if cmd.state is not None and cmd.state != '':
            self.view.log_state(cmd.state, cmd.ephemeral_context)
        else:
            self.view.log_info('actions/force command contains no state.')

        self.view.log_query(cmd.query, cmd.ephemeral_context)

        if self.view.controls.ignore_actions_force:
            self.view.log_system('Forced action ignored.')
            self.active_actions_force = None
            return

        # Check if all actions exist
        if not all(self.model.has_action(name) for name in cmd.action_names):
            self.view.log_warning('actions/force with invalid actions received. Discarding.\nInvalid actions: ' + ', '.join(name for name in cmd.action_names if not self.model.has_action(name)))
            self.active_actions_force = None
            return

        self.execute_actions_force(cmd)

    def on_action_result(self, cmd: ActionResultCommand):
        '''Handle the action/result command.'''

        self.view.log_system('Action result indicates ' + ('success' if cmd.success else 'failure'))

        self.view.log_debug(f'cmd.success: {cmd.success}, active_actions_force: {self.active_actions_force}')

        if not cmd.success and self.active_actions_force is not None:
            self.retry_actions_force(self.active_actions_force)
        else:
            self.active_actions_force = None

        if cmd.message is not None:
            self.view.log_action_result(cmd.success, cmd.message)
        elif cmd.success:
            self.view.log_info('Successful action result contains no message.')
        else:
            self.view.log_warning('Failed action result contains no message.')

        wx.CallAfter(self.view.on_action_result, cmd.success, cmd.message)

    def on_shutdown_ready(self, cmd: ShutdownReadyCommand):
        '''Handle the shutdown/ready command.'''

        self.view.log_warning('This command is not officially supported.')

    def on_unknown_command(self, json_cmd: Any):
        '''Handle an unknown command.'''

        # self.view.log_warning(f'Unknown command received: {json_cmd['command']}')

    def send_action(self, id: str, name: str, data: str | None):
        '''Send an action command to the API.'''

        self.view.log_system(f'Sending action: {name}')
        self.api.send_action(id, name, data)

        self.view.disable_actions() # Disable the actions until the result is received

    def send_actions_reregister_all(self):
        '''Send an actions/reregister_all command to the API.'''

        self.api.send_actions_reregister_all()

    def on_view_execute(self, action: NeuroAction) -> bool:
        '''
        Handle an action execution request from the view.
        Returns True if an action was sent, False if the action was cancelled.
        '''

        if not action.schema:
            self.send_action(next(self.id_generator), action.name, None) # No schema, so send the action immediately
            return True

        # If there is a schema, open a dialog to get the data
        result = self.view.show_action_dialog(action)
        if result is None:
            return False # User cancelled the dialog

        self.send_action(next(self.id_generator), action.name, result)
        return True

    def on_view_delete_action(self, name: str):
        '''Handle a request to delete an action from the view.'''

        self.model.remove_action_by_name(name)
        self.view.remove_action_by_name(name)

        self.view.log_system(f'Action deleted: {name}')

    def on_view_unlock(self):
        '''Handle a request to unlock the view.'''

        self.view.log_system('Unlocking actions.')
        self.view.enable_actions()

    def on_view_clear_logs(self):
        '''Handle a request to clear the logs from the view.'''

        self.view.clear_logs()
        self.view.log_system('Logs cleared.')

    def on_view_send_actions_reregister_all(self):
        '''Handle a request to send an actions/reregister_all command from the view.'''

        self.model.clear_actions()
        wx.CallAfter(self.view.clear_actions)
        self.send_actions_reregister_all()

    def on_view_send_shutdown_graceful(self):
        '''Handle a request to send a shutdown/graceful command with wants_shutdown=true from the view.'''

        self.api.send_shutdown_graceful(True)

    def on_view_send_shutdown_graceful_cancel(self):
        '''Handle a request to send a shutdown/graceful with wants_shutdown=false command from the view.'''

        self.api.send_shutdown_graceful(False)

    def on_view_send_shutdown_immediate(self):
        '''Handle a request to send a shutdown/immediate command from the view.'''

        self.api.send_shutdown_immediate()

    def execute_actions_force(self, cmd: ActionsForceCommand, retry: bool = False):
        self.active_actions_force = cmd

        if self.view.controls.auto_send:
            self.view.log_system('Automatically sending random action.')
            actions = [action for action in self.model.actions if action.name in cmd.action_names]
            action = random.choice(actions)

            if action.schema is None:
                self.send_action(next(self.id_generator), action.name, None)
            else:
                faker = JSF(action.schema)
                sample = faker.generate()
                self.send_action(next(self.id_generator), action.name, json.dumps(sample))

        else:
            wx.CallAfter(self.view.force_actions, cmd.state, cmd.query, cmd.ephemeral_context, cmd.action_names, retry)

    def retry_actions_force(self, cmd: ActionsForceCommand):
        '''Retry the actions/force command.'''

        if self.view.controls.ignore_actions_force:
            self.view.log_system('Forced action ignored.')
            self.active_actions_force = None
            return

        # Check if all actions exist
        if not all(self.model.has_action(name) for name in cmd.action_names):
            self.view.log_warning('Actions have been unregistered before retrying the forced action. Retry aborted.\nInvalid actions: ' + ', '.join(name for name in cmd.action_names if not self.model.has_action(name)))
            self.active_actions_force = None
            return

        self.view.log_system('Retrying forced action.')

        self.execute_actions_force(cmd, retry=True)



import json
from typing import Any, Callable, Optional
import jsonschema
import wx
from datetime import datetime as dt
from jsf import JSF

from .constants import VERSION
from .model import TonyModel, NeuroAction

#region Events

EVTTYPE_ADD_ACTION = wx.NewEventType()
EVT_ADD_ACTION = wx.PyEventBinder(EVTTYPE_ADD_ACTION, 1)

class AddActionEvent(wx.PyCommandEvent):
    '''An event for adding an action to the list.'''

    def __init__(self, id, action: NeuroAction):
        super().__init__(EVTTYPE_ADD_ACTION, id)
        self.action = action

EVTTYPE_ACTION_RESULT = wx.NewEventType()
EVT_ACTION_RESULT = wx.PyEventBinder(EVTTYPE_ACTION_RESULT, 1)

class ActionResultEvent(wx.PyCommandEvent):
    '''An event for an action result message.'''

    def __init__(self, id, success: bool, message: str | None):
        super().__init__(EVTTYPE_ACTION_RESULT, id)
        self.success = success
        self.message = message

EVT_TYPE_EXECUTE = wx.NewEventType()
EVT_EXECUTE = wx.PyEventBinder(EVT_TYPE_EXECUTE, 1)

class ExecuteEvent(wx.PyCommandEvent):
    '''An event for executing an action.'''

    def __init__(self, id, action: NeuroAction):
        super().__init__(EVT_TYPE_EXECUTE, id)
        self.action = action

#endregion

#region Constants

# Colors
LOG_COLOR_DEFAULT                       = wx.Colour(  0,   0,   0)
LOG_COLOR_TIMESTAMP                     = wx.Colour(  0, 128,   0)
LOG_COLOR_DEBUG                         = wx.Colour(128, 128, 128)
LOG_COLOR_INFO                          = wx.Colour(128, 192, 255)
LOG_COLOR_WARNING                       = wx.Colour(255, 192,   0)
LOG_COLOR_ERROR                         = wx.Colour(255,   0,   0)
LOG_COLOR_CRITICAL                      = wx.Colour(192,   0,   0)
LOG_COLOR_CONTEXT                       = LOG_COLOR_DEFAULT
LOG_COLOR_CONTEXT_QUERY                 = wx.Colour(255, 128, 255)
LOG_COLOR_CONTEXT_STATE                 = wx.Colour(128, 255, 128)
LOG_COLOR_CONTEXT_SILENT                = wx.Colour(128, 128, 128)
LOG_COLOR_CONTEXT_EPHEMERAL             = wx.Colour(128, 192, 255)
LOG_COLOR_CONTEXT_ACTION                = LOG_COLOR_DEFAULT
LOG_COLOR_CONTEXT_ACTION_RESULT_SUCCESS = wx.Colour(  0, 128,   0)
LOG_COLOR_CONTEXT_ACTION_RESULT_FAILURE = wx.Colour(255,   0,   0)
LOG_COLOR_RAW_INCOMING              = wx.Colour(  0,   0, 255)
LOG_COLOR_RAW_OUTGOING              = wx.Colour(255, 128, 192)

UI_COLOR_ERROR = wx.Colour(255, 192, 192)

LOG_LEVELS = {
    'DEBUG': 10,
    'INFO': 20,
    'WARNING': 30,
    'ERROR': 40,
    'CRITICAL': 50,
    'SYSTEM': 60,
}

#endregion

class TonyView:
    '''The view class for Tony.'''

    def __init__(self, app: wx.App, model: TonyModel, log_level: str):
        self.model = model

        self.controls = Controls()
        self.controls.set_log_level(log_level)

        self.frame = MainFrame(self)
        app.SetTopWindow(self.frame)

        self.action_dialog: Optional[ActionDialog] = None

        # Dependency injection
        self.on_execute: Callable[[NeuroAction], bool] = lambda action: None
        self.on_delete_action: Callable[[str], None] = lambda name: None
        self.on_unlock: Callable[[], None] = lambda: None
        self.on_clear_logs: Callable[[], None] = lambda: None
        self.on_send_actions_reregister_all: Callable[[], None] = lambda: None
        self.on_send_shutdown_graceful: Callable[[], None] = lambda: None
        self.on_send_shutdown_graceful_cancel: Callable[[], None] = lambda: None
        self.on_send_shutdown_immediate: Callable[[], None] = lambda: None

    def show(self):
        '''Show the main frame.'''

        self.frame.Show()

    def log_system(self, message: str):
        '''Log a command.'''

        if self.controls.get_log_level() <= LOG_LEVELS['SYSTEM']:
            self.frame.panel.log_notebook.log_panel.log(message)

    def log_debug(self, message: str):
        '''Log a debug message.'''

        if self.controls.get_log_level() <= LOG_LEVELS['DEBUG']:
            self.frame.panel.log_notebook.log_panel.log(message, 'Debug', LOG_COLOR_DEBUG)

    def log_info(self, message: str):
        '''Log an informational message.'''

        if self.controls.get_log_level() <= LOG_LEVELS['INFO']:
            self.frame.panel.log_notebook.log_panel.log(message, 'Info', LOG_COLOR_INFO)

    def log_warning(self, message: str):
        '''Log a warning message.'''

        if self.controls.get_log_level() <= LOG_LEVELS['WARNING']:
            self.frame.panel.log_notebook.log_panel.log(message, 'Warning', LOG_COLOR_WARNING)

    def log_error(self, message: str):
        '''Log an error message.'''

        if self.controls.get_log_level() <= LOG_LEVELS['ERROR']:
            self.frame.panel.log_notebook.log_panel.log(message, 'Error', LOG_COLOR_ERROR)

    def log_critical(self, message: str):
        '''Log a critical error message.'''

        if self.controls.get_log_level() <= LOG_LEVELS['CRITICAL']:
            self.frame.panel.log_notebook.log_panel.log(message, 'Critical', LOG_COLOR_CRITICAL)

    def log_context(self, message: str, silent: bool = False):
        '''Log a context message.'''

        tags = ['Context']
        colors = [LOG_COLOR_CONTEXT]

        if silent:
            tags.append('silent')
            colors.append(LOG_COLOR_CONTEXT_SILENT)

        self.frame.panel.log_notebook.context_log_panel.log(message, tags, colors)

    def log_description(self, message: str):
        '''Log an action description.'''

        self.frame.panel.log_notebook.context_log_panel.log(message, 'Action', LOG_COLOR_CONTEXT)

    def log_query(self, message: str, ephemeral: bool = False):
        '''Log an actions/force query.'''

        tags = ['Query']
        colors = [LOG_COLOR_CONTEXT_QUERY]

        if ephemeral:
            tags.append('ephemeral')
            colors.append(LOG_COLOR_CONTEXT_EPHEMERAL)

        self.frame.panel.log_notebook.context_log_panel.log(message, tags, colors)

    def log_state(self, message: str, ephemeral: bool = False):
        '''Log an actions/force state.'''

        tags = ['State']
        colors = [LOG_COLOR_CONTEXT_STATE]

        if ephemeral:
            tags.append('Ephemeral')
            colors.append(LOG_COLOR_CONTEXT_EPHEMERAL)

        self.frame.panel.log_notebook.context_log_panel.log(message, tags, colors)

    def log_action_result(self, success: bool, message: str | None):
        '''Log an action result message.'''

        
        self.frame.panel.log_notebook.context_log_panel.log(message, 'Result', LOG_COLOR_CONTEXT_ACTION_RESULT_SUCCESS if success else LOG_COLOR_CONTEXT_ACTION_RESULT_FAILURE)

    def log_raw(self, message: str, incoming: bool):
        '''Log raw data.'''

        tag = 'Game --> Neuro' if incoming else 'Game <-- Neuro'
        color = LOG_COLOR_RAW_INCOMING if incoming else LOG_COLOR_RAW_OUTGOING

        self.frame.panel.log_notebook.raw_log_panel.log(message, tag, color)

    def clear_logs(self):
        '''Clear all logs.'''

        self.frame.panel.log_notebook.log_panel.text.Clear()
        self.frame.panel.log_notebook.context_log_panel.text.Clear()
        self.frame.panel.log_notebook.raw_log_panel.text.Clear()

    def show_action_dialog(self, action: NeuroAction) -> Optional[str]:
        '''Show a dialog for an action. Returns the JSON string the user entered if "Send" was clicked, otherwise None.'''

        self.action_dialog = ActionDialog(self.frame, action, self.controls.validate_schema)
        result = self.action_dialog.ShowModal()
        text = self.action_dialog.text.GetValue()
        self.action_dialog.Destroy()
        self.action_dialog = None

        if result == wx.ID_OK:
            return text
        else:
            return None

    def close_action_dialog(self):
        '''
        Close the currently opened action dialog.
        Does nothing if no dialog is open.
        Handled as if the "Cancel" button was clicked.
        '''

        if self.action_dialog is not None:
            self.action_dialog.EndModal(wx.ID_CANCEL)
            self.action_dialog = None

    def add_action(self, action: NeuroAction):
        '''Add an action to the list.'''

        self.frame.panel.action_list.add_action(action)

    def remove_action_by_name(self, name: str):
        '''Remove an action from the list by name.'''

        self.frame.panel.action_list.remove_action_by_name(name)

    def enable_actions(self):
        '''Enable all action buttons.'''
        
        wx.CallAfter(self.frame.panel.action_list.execute_button.Enable)

    def disable_actions(self):
        '''Disable all action buttons.'''
        
        self.frame.panel.action_list.execute_button.Disable()

    def force_actions(self, state: str, query: str, ephemeral_context: bool, action_names: list[str], retry: bool = False):
        '''Show a dialog for forcing actions.'''

        actions = [action for action in self.model.actions if action.name in action_names]
        actions_force_dialog = ActionsForceDialog(self.frame, self, state, query, ephemeral_context, actions, retry)
        result = actions_force_dialog.ShowModal()
        actions_force_dialog.Destroy()

        # Executing the action has already been handled by the dialog
        if result != wx.ID_OK:
            self.log_system('Manually ignored forced action.')

    def clear_actions(self):
        '''Clear the list of actions.'''

        self.frame.panel.action_list.clear()

    def on_action_result(self, success: bool, message: str | None):
        '''
        Handle an action/result message.
        Enables the execute button.
        '''

        self.enable_actions()

class MainFrame(wx.Frame):
    '''The main frame for the Tony.'''

    def __init__(self, view: TonyView):
        super().__init__(None, title='Tony v' + VERSION)

        self.view = view
        self.panel = MainPanel(self)
        
        best_size: wx.Size = self.panel.GetBestSize()
        self.SetSize((850, 600))

class MainPanel(wx.Panel):
    '''The main panel for Tony.'''

    def __init__(self, parent):
        super().__init__(parent)

        self.action_list = ActionList(self, True)
        right_panel = wx.Panel(self)
        self.log_notebook = LogNotebook(right_panel)
        self.control_panel = ControlPanel(right_panel)

        right_panel_sizer = wx.BoxSizer(wx.VERTICAL)
        right_panel_sizer.Add(self.log_notebook, 1, wx.EXPAND | wx.ALL, 5)
        right_panel_sizer.Add(self.control_panel, 0, wx.EXPAND | wx.ALL, 5)
        right_panel.SetSizer(right_panel_sizer)

        self.sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.sizer.Add(self.action_list, 1, wx.EXPAND | wx.ALL, 5)
        self.sizer.Add(right_panel, 1, wx.EXPAND)
        self.SetSizer(self.sizer)
            
class ActionList(wx.Panel):
    '''The list of actions.'''
    
    def __init__(self, parent, can_delete: bool):
        super().__init__(parent, style=wx.BORDER_SUNKEN)

        self.actions: list[NeuroAction] = []

        self.list = wx.ListCtrl(self, style=wx.LC_REPORT | wx.LC_SINGLE_SEL)
        button_panel = wx.Panel(self)
        self.execute_button = wx.Button(button_panel, label='Execute')
        self.delete_button = wx.Button(button_panel, label='Delete')
        self.unlock_button = wx.Button(button_panel, label='Unlock')

        button_panel_sizer = wx.BoxSizer(wx.HORIZONTAL)
        button_panel_sizer.Add(self.execute_button, 0, wx.EXPAND | wx.ALL, 5)
        button_panel_sizer.Add(self.delete_button, 0, wx.EXPAND | wx.ALL, 5)
        button_panel_sizer.Add(self.unlock_button, 0, wx.EXPAND | wx.ALL, 5)
        button_panel.SetSizer(button_panel_sizer)

        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.sizer.Add(self.list, 1, wx.EXPAND | wx.ALL, 5)
        self.sizer.Add(button_panel, 0, wx.EXPAND)
        self.SetSizer(self.sizer)

        self.Bind(wx.EVT_BUTTON, self.on_execute, self.execute_button)
        self.Bind(wx.EVT_BUTTON, self.on_delete, self.delete_button)
        self.Bind(wx.EVT_BUTTON, self.on_unlock, self.unlock_button)

        self.list.InsertColumn(0, 'Name', width=150)
        self.list.InsertColumn(1, 'Description', width=240)
        self.list.InsertColumn(2, 'Schema', width=60)

        if not can_delete:
            self.delete_button.Disable()

    def add_action(self, action: NeuroAction):
        '''Add an action panel to the list.'''

        self.actions.append(action)

        self.list.Append([action.name, action.description, 'Yes' if action.schema is not None and action.schema != {} else 'No'])

    def remove_action_by_name(self, name: str):
        '''Remove an action panel from the list.'''

        self.actions = [action for action in self.actions if action.name != name]
        
        index = self.list.FindItem(-1, name)
        if(index != -1):
            self.list.DeleteItem(index)
        # else:
        #     self.GetTopLevelParent().view.log_error(f'Action "{name}" not found in list.')
    
    def clear(self):
        '''Clear the list of actions.'''

        self.actions.clear()
        self.list.DeleteAllItems()

    def on_execute(self, event: wx.CommandEvent):
        event.Skip()

        index = self.list.GetFirstSelected()

        if index == -1:
            return
        
        action = self.actions[index]

        top: MainFrame = self.GetTopLevelParent()
        sent = top.view.on_execute(action)
        
        if sent:
            self.GetEventHandler().ProcessEvent(ExecuteEvent(self.GetId(), action))
        top.view.log_debug(f'Sent: {sent}')

    def on_delete(self, event: wx.CommandEvent):
        event.Skip()

        index = self.list.GetFirstSelected()

        if index == -1:
            return
        
        action: NeuroAction = self.actions[index]

        top: MainFrame = self.GetTopLevelParent()
        top.view.on_delete_action(action.name)

    def on_unlock(self, event: wx.CommandEvent):
        event.Skip()

        top: MainFrame = self.GetTopLevelParent()
        top.view.on_unlock()
    
class LogNotebook(wx.Notebook):
    '''The notebook for logging messages.'''

    def __init__(self, parent):
        super().__init__(parent)

        self.log_panel = LogPanel(self)
        self.context_log_panel = LogPanel(self)
        self.raw_log_panel = LogPanel(self, text_ctrl_style=wx.TE_MULTILINE | wx.TE_READONLY | wx.TE_RICH | wx.HSCROLL)

        self.AddPage(self.log_panel, 'Log')
        self.AddPage(self.context_log_panel, 'Context')
        self.AddPage(self.raw_log_panel, 'Raw')

class LogPanel(wx.Panel):
    '''The panel for logging messages.'''

    def __init__(self, parent, text_ctrl_style = wx.TE_MULTILINE | wx.TE_READONLY | wx.TE_RICH):
        super().__init__(parent, style=wx.BORDER_SUNKEN)

        self.text = wx.TextCtrl(self, style=text_ctrl_style)
        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.sizer.Add(self.text, 1, wx.EXPAND)
        self.SetSizer(self.sizer)

    def log(self, message: str, tags: str | list[str] | None = [], tag_colors: wx.Colour | list[wx.Colour] | None = []):
        '''Log a message with optional tags and colors.'''

        # Convert single tags and colors to lists
        if isinstance(tags, str):
            tags = [tags]
        if isinstance(tag_colors, wx.Colour):
            tag_colors = [tag_colors]

        # Convert None to empty lists
        tags = tags or []
        tag_colors = tag_colors or []

        # Add default color for tags without color
        tag_colors += [LOG_COLOR_DEFAULT] * (len(tags) - len(tag_colors))

        # Log timestamp
        self.text.SetDefaultStyle(wx.TextAttr(LOG_COLOR_TIMESTAMP))
        self.text.AppendText(f'[{dt.now().strftime("%X")}] ')
        
        # Log tags
        for tag, tag_color in zip(tags, tag_colors):
            self.text.SetDefaultStyle(wx.TextAttr(tag_color))
            self.text.AppendText(f'[{tag}] ')
        
        # Log message
        self.text.SetDefaultStyle(wx.TextAttr(LOG_COLOR_DEFAULT))
        self.text.AppendText(f'{message}\n')

class ControlPanel(wx.Panel):
    '''The panel for controlling the application.'''

    def __init__(self, parent):
        super().__init__(parent, style=wx.BORDER_SUNKEN)

        self.view: TonyView = self.GetTopLevelParent().view

        # Create controls

        self.validate_schema_checkbox = wx.CheckBox(self, label='Validate JSON schema')
        self.ignore_actions_force_checkbox = wx.CheckBox(self, label='Ignore forced actions')
        self.auto_send_checkbox = wx.CheckBox(self, label='Automatically answer forced actions')

        latency_panel = wx.Panel(self)
        latency_text1 = wx.StaticText(latency_panel, label='L*tency:')
        self.latency_input = wx.TextCtrl(latency_panel, value='0', size=(50, -1))
        latency_text2 = wx.StaticText(latency_panel, label='ms')

        log_level_panel = wx.Panel(self)
        log_level_text = wx.StaticText(log_level_panel, label='Log level:')
        self.log_level_choice = wx.Choice(log_level_panel, choices=[s.capitalize() for s in LOG_LEVELS.keys()])

        self.clear_logs_button = wx.Button(self, label='Clear logs')
        self.send_actions_reregister_all_button = wx.Button(self, label='Clear all actions and request reregistration (experimental)')
        self.send_shutdown_graceful_button = wx.Button(self, label='Request graceful shutdown (experimental)')
        self.send_shutdown_graceful_cancel_button = wx.Button(self, label='Cancel graceful shutdown (experimental)')
        self.send_shutdown_immidiate_button = wx.Button(self, label='Request immediate shutdown (experimental)')

        # Create sizers

        latency_panel_sizer = wx.BoxSizer(wx.HORIZONTAL)
        latency_panel_sizer.Add(latency_text1, 0, wx.ALL | wx.ALIGN_CENTER, 2)
        latency_panel_sizer.Add(self.latency_input, 0, wx.ALL | wx.ALIGN_CENTER, 2)
        latency_panel_sizer.Add(latency_text2, 0, wx.ALL | wx.ALIGN_CENTER, 2)
        latency_panel.SetSizer(latency_panel_sizer)

        log_lever_panel_sizer = wx.BoxSizer(wx.HORIZONTAL)
        log_lever_panel_sizer.Add(log_level_text, 0, wx.ALL | wx.ALIGN_CENTER, 2)
        log_lever_panel_sizer.Add(self.log_level_choice, 0, wx.ALL | wx.ALIGN_CENTER, 2)
        log_level_panel.SetSizer(log_lever_panel_sizer)

        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.sizer.Add(self.validate_schema_checkbox, 0, wx.EXPAND | wx.ALL, 2)
        self.sizer.Add(self.ignore_actions_force_checkbox, 0, wx.EXPAND | wx.ALL, 2)
        self.sizer.Add(self.auto_send_checkbox, 0, wx.EXPAND | wx.ALL, 2)
        self.sizer.Add(latency_panel, 0, wx.EXPAND, 0)
        self.sizer.Add(log_level_panel, 0, wx.EXPAND, 0)
        self.sizer.Add(self.clear_logs_button, 0, wx.EXPAND | wx.ALL, 2)
        self.sizer.Add(self.send_actions_reregister_all_button, 0, wx.EXPAND | wx.ALL, 2)
        self.sizer.Add(self.send_shutdown_graceful_button, 0, wx.EXPAND | wx.ALL, 2)
        self.sizer.Add(self.send_shutdown_graceful_cancel_button, 0, wx.EXPAND | wx.ALL, 2)
        self.sizer.Add(self.send_shutdown_immidiate_button, 0, wx.EXPAND | wx.ALL, 2)
        self.SetSizer(self.sizer)

        # Bind events

        self.Bind(wx.EVT_CHECKBOX, self.on_validate_schema, self.validate_schema_checkbox)
        self.Bind(wx.EVT_CHECKBOX, self.on_ignore_actions_force, self.ignore_actions_force_checkbox)
        self.Bind(wx.EVT_CHECKBOX, self.on_auto_send, self.auto_send_checkbox)

        self.Bind(wx.EVT_TEXT, self.on_latency, self.latency_input)

        self.Bind(wx.EVT_CHOICE, self.on_log_level, self.log_level_choice)

        self.Bind(wx.EVT_BUTTON, self.on_clear_logs, self.clear_logs_button)
        self.Bind(wx.EVT_BUTTON, self.on_send_actions_reregister_all, self.send_actions_reregister_all_button)
        self.Bind(wx.EVT_BUTTON, self.on_send_shutdown_graceful, self.send_shutdown_graceful_button)
        self.Bind(wx.EVT_BUTTON, self.on_send_shutdown_graceful_cancel, self.send_shutdown_graceful_cancel_button)
        self.Bind(wx.EVT_BUTTON, self.on_send_shutdown_immediate, self.send_shutdown_immidiate_button)

        # Set default values

        self.validate_schema_checkbox.SetValue(True)
        self.ignore_actions_force_checkbox.SetValue(False)
        self.auto_send_checkbox.SetValue(False)
        # self.latency_input.SetValue('0')
        self.log_level_choice.SetStringSelection(self.view.controls.get_log_level_str())

        # Modify

    def on_clear_logs(self, event: wx.CommandEvent):
        event.Skip()

        self.view.on_clear_logs()

    def on_validate_schema(self, event: wx.CommandEvent):
        event.Skip()

        self.view.controls.validate_schema = event.IsChecked()

    def on_ignore_actions_force(self, event: wx.CommandEvent):
        event.Skip()

        self.view.controls.ignore_actions_force = event.IsChecked()

    def on_auto_send(self, event: wx.CommandEvent):
        event.Skip()

        self.view.controls.auto_send = event.IsChecked()

    def on_latency(self, event: wx.CommandEvent):
        event.Skip()

        try:
            latency = int(self.latency_input.GetValue())
            if latency < 0:
                raise ValueError('Latency must be non-negative.')
            elif latency > 10000:
                raise ValueError('Latency must not exceed 10000 ms.')
            self.view.controls.latency = latency
            self.latency_input.UnsetToolTip()
            self.latency_input.SetBackgroundColour(wx.NullColour) # Default color
        except ValueError as e:
            self.latency_input.SetToolTip(str(e))
            self.latency_input.SetBackgroundColour(UI_COLOR_ERROR)
        self.latency_input.Refresh()

    def on_log_level(self, event: wx.CommandEvent):
        event.Skip()

        sel = self.log_level_choice.GetSelection()
        self.view.controls.set_log_level(self.log_level_choice.GetString(sel))

    def on_send_actions_reregister_all(self, event: wx.CommandEvent):
        event.Skip()

        self.view.on_send_actions_reregister_all()

    def on_send_shutdown_graceful(self, event: wx.CommandEvent):
        event.Skip()

        self.view.on_send_shutdown_graceful()

    def on_send_shutdown_graceful_cancel(self, event: wx.CommandEvent):
        event.Skip()

        self.view.on_send_shutdown_graceful_cancel()

    def on_send_shutdown_immediate(self, event: wx.CommandEvent):
        event.Skip()

        self.view.on_send_shutdown_immediate()

class ActionDialog(wx.Dialog):

    def __init__(self, parent, action: NeuroAction, do_validate: bool):
        super().__init__(parent, title=action.name)

        self.action = action
        self.do_validate = do_validate

        self.text = wx.TextCtrl(self, style=wx.TE_MULTILINE)
        self.error_label = wx.StaticText(self, label='')
        button_panel = wx.Panel(self)
        self.send_button = wx.Button(button_panel, label='Send')
        self.show_schema_button = wx.Button(button_panel, label='Show Schema')
        self.cancel_button = wx.Button(button_panel, label='Cancel')

        button_panel_sizer = wx.BoxSizer(wx.HORIZONTAL)
        button_panel_sizer.Add(self.send_button, 0, wx.ALL, 2)
        button_panel_sizer.Add(self.show_schema_button, 0, wx.ALL, 2)
        button_panel_sizer.Add(self.cancel_button, 0, wx.ALL, 2)
        button_panel.SetSizer(button_panel_sizer)

        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.sizer.Add(self.text, 1, wx.EXPAND | wx.ALL, 2)
        self.sizer.Add(self.error_label, 0, wx.EXPAND | wx.ALL, 2)
        self.sizer.Add(button_panel, 0, wx.EXPAND)
        self.SetSizer(self.sizer)

        self.Bind(wx.EVT_BUTTON, self.on_send, self.send_button)
        self.Bind(wx.EVT_BUTTON, self.on_show_schema, self.show_schema_button)
        self.Bind(wx.EVT_BUTTON, self.on_cancel, self.cancel_button)

        faker = JSF(action.schema)
        sample = faker.generate()

        self.text.SetValue(json.dumps(sample, indent=2))

    def on_send(self, event: wx.CommandEvent):
        event.Skip()

        try:
            json_str = self.text.GetValue()
            json_cmd = json.loads(json_str)
            if self.do_validate:
                jsonschema.validate(json_cmd, self.action.schema)
            
            self.EndModal(wx.ID_OK)
            return
        
        except Exception as e:
            if isinstance(e, jsonschema.ValidationError):
                wx.MessageBox(f'JSON schema validation error: {e}', 'Error', wx.OK | wx.ICON_ERROR)
            elif isinstance(e, json.JSONDecodeError):
                wx.MessageBox(f'JSON decode error: {e}', 'Error', wx.OK | wx.ICON_ERROR)
            else:
                raise e

    def on_show_schema(self, event: wx.CommandEvent):
        event.Skip()

        wx.MessageBox(json.dumps(self.action.schema, indent=2), 'Schema', wx.OK | wx.ICON_INFORMATION)

    def on_cancel(self, event: wx.CommandEvent):
        event.Skip()

        self.EndModal(wx.ID_CANCEL)

class ActionsForceDialog(wx.Dialog):

    def __init__(self, parent, view: TonyView, state: str, query: str, ephemeral_context: bool, actions: list[NeuroAction], retry: bool = False):
        title = 'Forced Action' if not retry else 'Retry Forced Action'
        super().__init__(parent, title=title, style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER)

        self.view = view
        self.state = state
        self.query = query
        self.ephemeral_context = ephemeral_context
        self.actions = actions

        self.state_label = wx.StaticText(self, label=f'State: {state}')
        self.query_label = wx.StaticText(self, label=f'Query: {query}')
        self.ephemeral_context_label = wx.StaticText(self, label=f'Ephemeral Context: {ephemeral_context}')
        self.action_list = ActionList(self, False)

        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.sizer.Add(self.state_label, 0, wx.EXPAND | wx.ALL, 2)
        self.sizer.Add(self.query_label, 0, wx.EXPAND | wx.ALL, 2)
        self.sizer.Add(self.ephemeral_context_label, 0, wx.EXPAND | wx.ALL, 2)
        self.sizer.Add(self.action_list, 1, wx.EXPAND | wx.ALL, 2)
        self.SetSizer(self.sizer)

        for action in actions:
            self.action_list.add_action(action)

        self.action_list.list.Select(0)

        self.Bind(EVT_EXECUTE, self.on_execute, self.action_list)

    def on_execute(self, event: ExecuteEvent):
        event.Skip()

        self.EndModal(wx.ID_OK)

class Controls:
    '''The content of the control panel.'''

    def __init__(self):
        self.validate_schema: bool = True
        self.ignore_actions_force: bool = False
        self.auto_send: bool = False
        self.latency: int = 0

        self.__log_level_str: str = 'INFO'
        self.__log_level: int = LOG_LEVELS['INFO']

    def set_log_level(self, log_level: str):
        '''Set the log level.'''

        self.__log_level_str = log_level
        self.__log_level = LOG_LEVELS[log_level]

    def get_log_level(self) -> int:
        '''Get the log level.'''

        return self.__log_level
    
    def get_log_level_str(self) -> str:
        '''Get the log level as a string.'''

        return self.__log_level_str
    


from typing import Any, Optional

class TonyModel:
    
    def __init__(self):
        self.actions: list[NeuroAction] = []

    def add_action(self, action: 'NeuroAction'):
        '''Add an action to the list.'''
        
        self.actions.append(action)

    def remove_action(self, action: 'NeuroAction'):
        '''Remove an action from the list.'''
        
        self.actions.remove(action)
        pass

    def remove_action_by_name(self, name: str):
        '''Remove an action from the list by name.'''
        
        self.actions = [action for action in self.actions if action.name != name]
        pass

    def clear_actions(self):
        '''Clear all actions from the list.'''
        
        self.actions.clear()
        pass

    def has_action(self, name: str) -> bool:
        '''Check if an action exists in the list.'''
        
        return any(action.name == name for action in self.actions)
    
    def get_action_by_name(self, name: str) -> Optional['NeuroAction']:
        '''Get an action by name.'''
        
        for action in self.actions:
            if action.name == name:
                return action
        return None

class NeuroAction:
    
    def __init__(self, name: str, description: str, schema: Optional[dict[str, Any]]):
        self.name = name
        self.description = description
        self.schema = schema



import logging
import wx
import sys
from getopt import getopt
from git import CommandError, GitCommandError, Repo
from git.exc import InvalidGitRepositoryError
import subprocess

from .controller import TonyController
from .constants import APP_NAME, VERSION, GIT_REPO_URL

help_message = '''
Usage: python -m src [options]

Options:
    -h, --help:
        Show this help message and exit.

    -a, --addr, --address:
        The address to start the websocket server on. Default is localhost.

    -l, --log, --log-level:
        The log level to use. Default is INFO. Must be one of: DEBUG, INFO,
        WARNING, ERROR, SYSTEM.

    -p, --port:
        The port number to start the websocket server on. Default is 8000.

    --update:
        Update the program to the latest version, if available. Only works if
        the program is in a git repository.
    
    -v, --version:
        Show the version of the program and exit.
'''


def cli_run() -> None:
    options, _ = getopt(sys.argv[1:], 'ha:l:p:v', ['help', 'addr=', 'address=', 'log=', 'log-level=', 'port=', 'update', 'version'])

    address = 'localhost'
    port = 8000
    log_level = 'INFO'

    for option, value in options:
        match option:
            case '-h' | '--help':
                print(help_message)
                sys.exit(0)

            case '-a' | '--addr' | '--address':
                address = value

            case '-l' | '--log' | '--log-level':
                if value.upper() not in ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'SYSTEM']:
                    print('Invalid log level. Must be one of: DEBUG, INFO, WARNING, ERROR, SYSTEM.')
                    sys.exit(1)
                log_level = value.upper()

            case '-p' | '--port':
                port = int(value)

            case '--update':
                try:
                    repo = Repo('.')

                    print('Checking for updates...')

                    repo.remote().fetch()

                    if repo.head.commit == repo.remote().refs.master.commit: # Check if the local commit is the same as the remote commit
                        print('Program is already up to date.')
                        sys.exit(0)

                    print('Pulling changes from remote repository...')

                    # Only allow fast-forward merges so nothing breaks if the program is modified
                    repo.remote().pull(ff_only=True)

                    if repo.head.commit != repo.remote().refs.master.commit: # Check if the local commit is still different from the remote commit
                        print('Failed to update program.')
                        print('Please update manually using git or reinstall the program from ' + GIT_REPO_URL + '.')
                        sys.exit(1)

                    # Install dependencies
                    print('Installing dependencies...')
                    subprocess.check_call([sys.executable, '-m', 'pip', 'install', '-r', 'requirements.txt'])

                    print('Program updated successfully.')
                    sys.exit(0)
                
                except InvalidGitRepositoryError:
                    print('Program is not in a git repository.')
                    print('Please install the new version manually from ' + GIT_REPO_URL + '.')

                except CommandError as e:
                    print(e)
                    print()
                    print('Failed to update program.')
                    print('Please update manually using git or reinstall the program from ' + GIT_REPO_URL + '.')

                except subprocess.CalledProcessError as e:
                    print(e)
                    print()
                    print('Failed to install dependencies.')
                    print('Please install the dependencies manually using pip.')

                sys.exit(1)

            case '-v' | '--version':
                print(f'{APP_NAME} v{VERSION}')
                sys.exit(0)

    # Check if the program is up to date
    try:
        repo = Repo('.')
        repo.remote().fetch()

        if repo.head.commit != repo.remote().refs.master.commit: # Check if the local commit is different from the remote commit
            print('An update is available. Run "python -m src --update" to update.')

    except InvalidGitRepositoryError:
        print('Warning: Update checking is not yet implemented for PyPI distributions. Please check for updates manually until this feature is implemented.')

    except GitCommandError:
        print('Cannot check for updates. Please check your internet connection.')

    # Start the program
    app = wx.App()
    controller = TonyController(app, log_level)
    controller.run(address, port)


if __name__ == '__main__':
    cli_run()



from __future__ import annotations

import json
from threading import Lock
import traceback
from typing import TYPE_CHECKING, Any, NamedTuple, Final

import jsonschema
import jsonschema.exceptions
import trio
import wx
from trio_websocket import (
    ConnectionClosed,
    WebSocketConnection,
    WebSocketRequest,
    serve_websocket,
)

if TYPE_CHECKING:
    from outcome import Outcome
    from collections.abc import Callable

from .model import NeuroAction


ACTION_NAME_ALLOWED_CHARS: Final = 'abcdefghijklmnopqrstuvwxyz0123456789_-'

# See https://github.com/VedalAI/neuro-game-sdk/blob/main/API/SPECIFICATION.md#action
INVALID_SCHEMA_KEYS: Final = frozenset({"$anchor", "$comment", "$defs", "$dynamicAnchor", "$dynamicRef", "$id", "$ref", "$schema", "$vocabulary", "additionalProperties", "allOf", "anyOf", "contentEncoding", "contentMediaType", "contentSchema", "dependentRequired", "dependentSchemas", "deprecated", "description", "else", "if", "maxProperties", "minProperties", "not", "oneOf", "patternProperties", "readOnly", "then", "title", "unevaluatedItems", "unevaluatedProperties", "writeOnly"})


class NeuroAPI:

    def __init__(self) -> None:
        self.message_send_channel: trio.MemorySendChannel[str] | None = None
        self.queue_lock = Lock() # threading, not trio
        self.current_game = ''
        self.current_action_id: str | None = None

        # Dependency injection
        self.on_startup: Callable[[StartupCommand], None] = lambda cmd: None
        self.on_context: Callable[[ContextCommand], None] = lambda cmd: None
        self.on_actions_register: Callable[[ActionsRegisterCommand], None] = lambda cmd: None
        self.on_actions_unregister: Callable[[ActionsUnregisterCommand], None] = lambda cmd: None
        self.on_actions_force: Callable[[ActionsForceCommand], None] = lambda cmd: None
        self.on_action_result: Callable[[ActionResultCommand], None] = lambda cmd: None
        self.on_shutdown_ready: Callable[[ShutdownReadyCommand], None] = lambda cmd: None
        self.on_unknown_command: Callable[[Any], None] = lambda cmd: None
        self.log_system: Callable[[str], None] = lambda message: None
        self.log_debug: Callable[[str], None] = lambda message: None
        self.log_info: Callable[[str], None] = lambda message: None
        self.log_warning: Callable[[str], None] = lambda message: None
        self.log_error: Callable[[str], None] = lambda message: None
        self.log_critical: Callable[[str], None] = lambda message: None
        self.log_raw: Callable[[str, bool], None] = lambda message, incoming: None
        self.get_delay: Callable[[], float] = lambda: 0.0

        self.async_library_running = False
        self.async_library_root_cancel: trio.CancelScope

    def start(self, address: str, port: int) -> None:
        '''Start hosting the websocket server with Trio in the background.'''

        if self.async_library_running:
            # Already running, skip
            self.log_critical('Something attempted to start websocket server a 2nd time, ignoring.')
            return

        def done_callback(run_outcome: Outcome[None]) -> None:
            '''Called when trio run completes.'''

            assert self.async_library_running, 'How can stop running if not running?'
            self.async_library_running = False
            # Unwrap to make sure exceptions are printed
            run_outcome.unwrap()

        self.async_library_running = True
        self.async_library_root_cancel = trio.CancelScope()

        async def root_run() -> None:
            '''Root async run, wrapped with async_library_root_cancel so it's able to be stopped remotely.'''

            with self.async_library_root_cancel:
                await self._run(address, port)

        try:
            # Start the Trio guest run
            trio.lowlevel.start_guest_run(
                root_run,
                done_callback=done_callback,
                run_sync_soon_threadsafe=wx.CallAfter,
                host_uses_signal_set_wakeup_fd=False,
                restrict_keyboard_interrupt_to_checkpoints=True,
                strict_exception_groups=True,
            )
        except Exception:
            # Make sure async_library_running can never be in invalid state
            # even if trio fails to launch for some reason (shouldn't happen but still)
            self.async_library_running = False
            raise

    def stop(self) -> None:
        '''Stop hosting background websocket server.'''

        if not self.async_library_running:
            return
        self.async_library_root_cancel.cancel()

    async def _run(self, address: str, port: int) -> None:
        '''Server run root function.'''

        self.log_system(f'Starting websocket server on ws://{address}:{port}.')
        await serve_websocket(self._handle_websocket_request, address, port, ssl_context=None)

    @property
    def client_connected(self) -> bool:
        '''Is there a client connected?'''

        return self.message_send_channel is not None

    async def _handle_websocket_request(self, request: WebSocketRequest) -> None:
        '''Handle websocket connection request.'''

        if self.client_connected:
            # 503 is "service unavailable"
            await request.reject(503, body=b'Server does not support multiple connections at once currently')
            self.log_error('Another client attempted to connect, rejecting, server does not support multiple connections at once currently')
            return

        try:
            # With blocks for send and receive to handle closing when done
            # Using message_send_channel to send websocket messages synchronously
            # Zero here means no buffer, send not allowed to happen if receive channel has
            # not read prior message waiting yet.
            self.message_send_channel, receive_channel = trio.open_memory_channel[str](0)
            with (self.message_send_channel, receive_channel):
                # Accept connection
                async with await request.accept() as connection:
                    await self._handle_client_connection(connection, receive_channel)
        finally:
            self.message_send_channel = None

    async def _handle_client_connection(
        self,
        websocket: WebSocketConnection,
        receive_channel: trio.MemoryReceiveChannel[str],
    ) -> None:
        '''Handle websocket connection lifetime.'''

        try:
            async with trio.open_nursery() as nursery:
                # Start running connection read and write tasks in the background
                nursery.start_soon(self._handle_consumer, websocket, nursery.cancel_scope)
                nursery.start_soon(self._handle_producer, websocket, receive_channel)
        except trio.Cancelled:
            self.log_system('Closing current websocket connection.')

    async def _handle_consumer(
        self,
        websocket: WebSocketConnection,
        cancel_scope: trio.CancelScope,
    ) -> None:
        '''Handle websocket reading head.'''

        while True:
            try:
                # Read message from websocket
                message = await websocket.get_message()
            except ConnectionClosed:
                self.log_system('Websocket connection closed by client.')
                break

            try:
                json_cmd = json.loads(message)
                self.log_raw(json.dumps(json_cmd, indent=2), True)

                game = json_cmd.get('game', {})
                data = json_cmd.get('data', {})

                if game == '' or not isinstance(game, str):
                    self.log_warning('Game name is not set.')
                else:
                    # Check game name
                    if json_cmd['command'] == 'startup' or json_cmd['command'] == 'game/startup':
                        self.current_game = game
                    elif self.current_game != game:
                        self.log_warning('Game name does not match the current game.')
                    elif self.current_game == '':
                        self.log_warning('No startup command received.')

                # Check action result when waiting for it
                if self.current_action_id is not None and json_cmd['command'] == 'actions/force':
                    self.log_warning(f'Received actions/force while waiting for action/result.')

                self.log_system(f'Command received: {json_cmd["command"]}')

                # Handle the command
                match json_cmd['command']:
                    case 'startup' | 'game/startup':
                        self.current_action_id = None

                        self.on_startup(StartupCommand())

                        if json_cmd['command'] == 'game/startup':
                            self.log_warning('"game/startup" command is deprecated. Use "startup" instead.')

                    case 'context':
                        self.on_context(ContextCommand(data['message'], data['silent']))

                    case 'actions/register':
                        # Check the actions
                        for action in data['actions']:
                            # Check the schema
                            if 'schema' in action and action['schema'] != {}:
                                try:
                                    jsonschema.Draft7Validator.check_schema(action['schema'])
                                except jsonschema.exceptions.SchemaError as e:
                                    self.log_error(f'Invalid schema for action "{action["name"]}": {e}')
                                    continue

                                invalid_keys = self.check_invalid_keys_recursive(action['schema'])

                                if len(invalid_keys) > 0:
                                    self.log_warning(f'Disallowed keys in schema: {", ".join(invalid_keys)}')

                            # Check the name
                            if not isinstance(action['name'], str):
                                self.log_error(f'Action name is not a string: {action["name"]}')
                                continue

                            if not all(c in ACTION_NAME_ALLOWED_CHARS for c in action['name']):
                                self.log_warning(f'Action name is not a lowercase string.')

                            if action['name'] == '':
                                self.log_warning('Action name is empty.')

                        self.on_actions_register(ActionsRegisterCommand(data['actions']))

                    case 'actions/unregister':
                        self.on_actions_unregister(ActionsUnregisterCommand(data['action_names']))

                    case 'actions/force':
                        self.on_actions_force(ActionsForceCommand(data.get('state'), data['query'], data.get('ephemeral_context', False), data['action_names']))

                    case 'action/result':
                        # Check if an action/result was expected
                        if self.current_action_id is None:
                            self.log_warning('Unexpected action/result.')
                        # Check if the action ID matches
                        elif self.current_action_id != data['id']:
                            self.log_warning(f'Received action ID "{data["id"]}" does not match the expected action ID "{self.current_action_id}".')

                        self.log_debug(f'Action ID: {data["id"]}')

                        self.current_action_id = None
                        self.on_action_result(ActionResultCommand(data['success'], data.get('message', None)))

                    case 'shutdown/ready':
                        self.log_warning('This command is not officially supported.')
                        self.on_shutdown_ready(ShutdownReadyCommand())

                    case _:
                        self.log_warning('Unknown command.')
                        self.on_unknown_command(json_cmd)

            except Exception as exc:
                self.log_error(f'Error while handling message: {exc}')
                traceback.print_exception(exc)
                break
        # Cancel (stop) writing head
        cancel_scope.cancel()

    async def _handle_producer(
        self,
        websocket: WebSocketConnection,
        receive_channel: trio.MemoryReceiveChannel[str],
    ) -> None:
        '''Handle websocket writing head.'''

        while True:
            # Wait for messages from sending side of memory channel (queue)
            message = await receive_channel.receive()

            # Artificial latency
            # Make sure never < 0 or raises ValueError
            await trio.sleep(max(0, self.get_delay()))

            # Write message
            # If connection failure happens, will crash the read head
            # because both share a nursery, ensuring connection closes
            await websocket.send_message(message)

            try:
                self.log_raw(json.dumps(json.loads(message), indent=2), False)
            except json.JSONDecodeError:
                self.log_raw(message, False)

    def _submit_message(self, message: str) -> bool:
        '''Submit a message to the send queue. Return True if client connected.'''

        with self.queue_lock:
            if not self.client_connected:
                self.log_error('No client connected!')
                return False
            # Otherwise send channel should exist
            # type checkers need help understanding `self.client_connected`
            assert self.message_send_channel is not None
            self.message_send_channel.send_nowait(message)
        return True

    def send_action(self, id_: str, name: str, data: str | None) -> bool:
        '''Send an action command. Return True if actually sent.'''

        obj = {
            'command': 'action',
            'data': {
                'id': id_,
                'name': name,
                'data': data,
            }
        }

        message = json.dumps(obj)

        if not self._submit_message(message):
            return False

        self.current_action_id = id_
        self.log_system('Command sent: action')
        self.log_debug(f'Action ID: {id_}')

        return True

    def send_actions_reregister_all(self) -> bool:
        '''Send an actions/reregister_all command.'''

        message = json.dumps({
            'command': 'actions/reregister_all'
        })

        if not self._submit_message(message):
            return False

        self.log_system('Command sent: actions/reregister_all')
        self.log_warning('This command is not officially supported.')

        return True

    def send_shutdown_graceful(self, wants_shutdown: bool) -> bool:
        '''Send a shutdown/graceful command.'''

        message = json.dumps({
            'command': 'shutdown/graceful',
            'data': {
                'wants_shutdown': wants_shutdown
            }
        })

        if not self._submit_message(message):
            return False

        self.log_system('Command sent: shutdown/graceful')
        self.log_warning('This command is not officially supported.')

        return True

    def send_shutdown_immediate(self) -> bool:
        '''Send a shutdown/immediate command.'''

        message = json.dumps({
            'command': 'shutdown/immediate'
        })

        if not self._submit_message(message):
            return False

        self.log_system('Command sent: shutdown/immediate')
        self.log_warning('This command is not officially supported.')

        return True

    def check_invalid_keys_recursive(self, sub_schema: dict[str, Any]) -> list[str]:
        '''
        Recursively checks for invalid keys in the schema.
        Returns a list of invalid keys that were found.
        '''

        invalid_keys = []

        for key, value in sub_schema.items():
            if key in INVALID_SCHEMA_KEYS:
                invalid_keys.append(key)
            elif isinstance(value, str | int):
                pass
            elif isinstance(value, dict):
                invalid_keys.extend(self.check_invalid_keys_recursive(value))
            elif isinstance(value, list):
                for item in value:
                    if isinstance(item, dict):
                        invalid_keys.extend(self.check_invalid_keys_recursive(item))
            else:
                self.log_error(f'Unhandled schema key type {type(key)!r} ({key!r})')

        return invalid_keys


class StartupCommand:
    pass  # No data needed


class ContextCommand(NamedTuple):
    message: str
    silent: bool


class ActionsRegisterCommand:
    __slots__ = ('actions',)

    def __init__(self, actions: list[dict[str, Any]]) -> None:
        # 'schema' may be omitted, so get() is used
        self.actions = [
            NeuroAction(
                action['name'],
                action['description'],
                action.get('schema'),
            )
            for action in actions
        ]


class ActionsUnregisterCommand(NamedTuple):
    action_names: list[str]


class ActionsForceCommand(NamedTuple):
    state: str | None
    query: str
    ephemeral_context: bool
    action_names: list[str]


class ActionResultCommand(NamedTuple):
    success: bool
    message: str | None


class ShutdownReadyCommand:
    pass  # No data needed
